/*
 *  Copyright (c) 2010-2013, Rutgers, The State University of New Jersey
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions are met:
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above copyright
 *        notice, this list of conditions and the following disclaimer in the
 *        documentation and/or other materials provided with the distribution.
 *      * Neither the name of the organization(s) stated above nor the
 *        names of its contributors may be used to endorse or promote products
 *        derived from this software without specific prior written permission.
 *  
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *  ARE DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
 *  DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF 
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/**
 * @file   mfsocketmanager.cpp
 * @author wontoniii (bronzino@winlab.rutgers.edu)
 * @date   August, 2013
 * @brief  Class that handles the tresources for each socket.
 *
 * Class that handles the tresources for each socket.
 */

#include <arpa/inet.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <list>
#include <vector>
#include <queue>
#include <errno.h>
#include <unistd.h>

#include <include/mfroutingheader.hh>
#include <include/mfguid.hh>
#include <include/mfclientipc.h>

#include "mftypes.h"
#include "mfbuffer.h"
#include "mfchunkinfo.h"
#include "mfstackevent.h"
#include "mfserver.h"
#include "mftransport.h"
#include "mfsockettable.h"
#include "mfsocketmanager.h"
#include "mfpacketsupport.h"


//Add std size buffer
MF_SocketManager::MF_SocketManager() : buffer(10){
	ID = 0;
	open = false;
	lastGetRequest = 0;
}

//Select transport based on profile
MF_SocketManager::MF_SocketManager(MF_SocketTable *st, MF_EventQueue *eq, MF_Server *ms, u_int UID) : buffer(10){
	ID = UID;
	this->eq = eq;
	this->ms = ms;
	this->st = st;
	open = false;
	lastGetRequest = 0;
}

MF_SocketManager::MF_SocketManager(MF_SocketTable *st, MF_EventQueue *eq, MF_Server *ms, u_int UID, int pCount) : buffer(pCount) {
	MF_Log::mf_log(MF_DEBUG, "MF_SocketManager: initializing");
	ID = UID;
	this->eq = eq;
	this->ms = ms;
	this->st = st;
	open = false;
	lastGetRequest = 0;
}

bool MF_SocketManager::init(){
	cliAddr.sun_family = AF_LOCAL;
	sprintf(cliAddr.sun_path, "%s%010x", CLIENT_PATH, ID);
	mConnSock = socket(AF_LOCAL, SOCK_STREAM, 0);
	if (mConnSock < 0) {
		MF_Log::mf_log(MF_ERROR,"socket() error:%s\n",strerror(errno));
		//TODO: send error message to api layer
		return false;
	}
	struct sockaddr_un dataAddr;
	dataAddr.sun_family = AF_LOCAL;
	sprintf(dataAddr.sun_path, "%s%010x", SERVER_DATA_PATH, ID);
	unlink(dataAddr.sun_path);
	if (bind(mConnSock, (struct sockaddr *)&dataAddr, sizeof(dataAddr)) < 0 ){
		MF_Log::mf_log(MF_ERROR, "MF_SocketManager:MF_SocketManager() bind error constructing server datapath :%s",strerror(errno));
		close(mConnSock);
		//TODO: send error message to api layer
		return false;
	}
#ifdef __ANDROID__
	char command[128];
	strcpy(command, "chmod 777 ");
	strcat(command, dataAddr.sun_path);
	system(command);
#endif
	if (listen(mConnSock, 1) < 0) {
                MF_Log::mf_log(MF_ERROR, "MF_SocketManager:MF_SocketManager() listen error constructing server datapath :%s",strerror(errno));
                close(mConnSock);
		return false;
        }
	open = true;
	return true;
}

int MF_SocketManager::connectToClient(){
	if((mDataSock = accept(mConnSock, NULL, NULL)) < 0) {
		MF_Log::mf_log(MF_ERROR, "MF_SocketManager:MF_SocketManager() accept error constructing server datapath :%s",strerror(errno));
		return -1;
	}
	return 0;
}

bool MF_SocketManager::isOpen(){
	return open;
}

MF_SocketManager::~MF_SocketManager(){
	
}

u_int MF_SocketManager::getID(){
	return ID;
}

void MF_SocketManager::setID(u_int id){
	ID = id;
}

void MF_SocketManager::addGUID(int GUID){
	GUIDs.push_back(GUID);
}

void MF_SocketManager::removeGUID(int GUID){
	GUIDs.remove(GUID);
}

void MF_SocketManager::addTID(u_int TID){
	TIDs.push_back(TID);
}

void MF_SocketManager::removeTID(u_int TID){
	TIDs.remove(TID);
}

list<int>::iterator MF_SocketManager::getGUIDBegin(){
	return GUIDs.begin();
}

list<int>::iterator MF_SocketManager::getGUIDEnd(){
	return GUIDs.end();
}

list<u_int>::iterator MF_SocketManager::getTIDBegin(){
	return TIDs.begin();
}

list<u_int>::iterator MF_SocketManager::getTIDEnd(){
	return TIDs.end();
}

void MF_SocketManager::openReplyError(){
	MF_Log::mf_log(MF_DEBUG, "MF_SocketManager:openError a socket manager is already opened with UID %lu",ID);
	struct MsgOpenReply openR;
	openR.type = MSG_TYPE_OPEN_REPLY;
	openR.retValue = UID_EXISTS;
	if(ms->send((char *)(&openR), sizeof(struct MsgOpenReply), &cliAddr)<=0){
		MF_Log::mf_log(MF_ERROR, "MF_SocketManager:openSuccess error replying to API");
	}
}

void MF_SocketManager::openReplySuccess(){
	MF_Log::mf_log(MF_DEBUG, "MF_SocketManager:openSuccess a socket manager is opened with UID %lu",ID);
	struct MsgOpenReply openR;
	openR.type = MSG_TYPE_OPEN_REPLY;
	openR.retValue = 0;
	if(ms->send((char *)(&openR), sizeof(struct MsgOpenReply), &cliAddr)<=0){
		MF_Log::mf_log(MF_ERROR, "MF_SocketManager:openSuccess error replying to API");
	}
}

//Select transport based on profile and opts and set default GUID
int MF_SocketManager::openRequest(const char *profile, mfflag_t opts, const int GUID){
	MF_Log::mf_log(MF_DEBUG, "MF_SocketManager:open request analyizing request profile=%s opts=%lu guid=%d", profile, opts, GUID);
	if(!strcmp("basic", profile)){
		type = 1;
	}
	else if (!strcmp("content", profile)) {
		type = 2;
		//TODO: remove this to enable content
//		MF_Log::mf_log(MF_ERROR, "MF_SocketManager:openRequest wrong profile request %s", profile);
//		return -1;
	}
	else{
		MF_Log::mf_log(MF_ERROR, "MF_SocketManager:openRequest wrong profile request %s", profile);
		return -1;
	}
	open = true;
	st->addManager(this);
	//TODO select unique id for transport and select best transport given profile and opts
	//TODO TID should be assigned as system wise unique one
	trnsp = new MF_BasicTransport(ID, &buffer, eq);
	st->addTID(ID, ID);
	if(GUID!=0){
		GUIDs.push_back(GUID);
		st->addGUID(ID, GUID);
	}
	else{
		MF_Log::mf_log(MF_DEBUG, "MF_SocketManager:openRequest using device default srcGUID");
	}
	MF_Log::mf_log(MF_DEBUG, "MF_SocketManager:openRequest opened new socket with ID=%u default srcGUID=%d", ID, GUID);
	return 0;
}

//TODO: optimize the next two functions
int MF_SocketManager::recvSinglePacket(u_char *buf, int size, int offset){
	//MF_Log::mf_log(MF_DEBUG, "MF_SocketManager:recvSinglePacket receiving %d bytes with offset for buffer %d", size, offset);
	int rec = 0, totRec = 0;
	while(totRec!=size){
		rec = recv(mDataSock, buf + rec + offset, size - rec, 0);
		//MF_Log::mf_log(MF_DEBUG, "MF_SocketManager:recvSinglePacket received %d bytes", rec);
		if(rec<=0){
			MF_Log::mf_log(MF_ERROR, "MF_SocketManager:recvSinglePacket errno: %d: %s", errno, strerror(errno));
			return rec;
		}
		totRec+=rec;
	}
	return totRec;
}

//Receive data from api
int MF_SocketManager::recvData(vector<u_char *> *chkPkt, u_int chunkSize){
	MF_Log::mf_log(MF_DEBUG, "MF_SocketManager:recvData Receiving a message (size = %d) from app for ID %u", chunkSize, ID);
	int size = 0; //for debug only
	int recSize = 0; //for debug only
	int reminSize = (int)chunkSize;
	int i = 0;
	while(recSize!=(int)chunkSize){
		if ((reminSize <= (int)(MF_PacketSupport::MAX_PAYLOAD_SIZE - MF_PacketSupport::HIGH_HEADER_LEN)) && (i%MAX_CHUNK_SIZE == 0)) {
			size = recvSinglePacket((*chkPkt)[i], reminSize, MF_PacketSupport::TOTAL_HEADER_LEN);
			//MF_Log::mf_log(MF_DEBUG, "MF_SocketManager:recvData Last loop first chunk Receiving %d received %d", reminSize, size);
			recSize += size;
			if(size < 0) return -1;
		}
		else if ((reminSize > (int)(MF_PacketSupport::MF_PacketSupport::MAX_PAYLOAD_SIZE - MF_PacketSupport::HIGH_HEADER_LEN)) && (i%MAX_CHUNK_SIZE == 0)) {
			size = recvSinglePacket((*chkPkt)[i], MF_PacketSupport::MAX_PAYLOAD_SIZE - MF_PacketSupport::HIGH_HEADER_LEN, MF_PacketSupport::TOTAL_HEADER_LEN);
			//MF_Log::mf_log(MF_DEBUG, "MF_SocketManager:recvData First packet of chunk Receiving %d received %d", MF_PacketSupport::MAX_PAYLOAD_SIZE - MF_PacketSupport::HIGH_HEADER_LEN, size);
			recSize += size;
			reminSize = reminSize - (MF_PacketSupport::MAX_PAYLOAD_SIZE - MF_PacketSupport::HIGH_HEADER_LEN);
			i++;
			if(size < 0) return -1;
		}
		else if ((reminSize <= MF_PacketSupport::MAX_PAYLOAD_SIZE) && (i%MAX_CHUNK_SIZE != 0)) {
			size = recvSinglePacket((*chkPkt)[i], reminSize, MF_PacketSupport::LOW_HEADER_LEN);
			//MF_Log::mf_log(MF_DEBUG, "MF_SocketManager:recvData Last loop not first packet of chunk Receiving %d received %d", reminSize, size);
			recSize += size;
			if(size < 0) return -1;
		}
		else if ((reminSize > MF_PacketSupport::MAX_PAYLOAD_SIZE) && (i%MAX_CHUNK_SIZE != 0)) {
			size = recvSinglePacket((*chkPkt)[i], MF_PacketSupport::MAX_PAYLOAD_SIZE, MF_PacketSupport::LOW_HEADER_LEN);
			//MF_Log::mf_log(MF_DEBUG, "MF_SocketManager:recvData not first Receiving %d received %d", MF_PacketSupport::MAX_PAYLOAD_SIZE, size);
			recSize += size;
			reminSize = reminSize - MF_PacketSupport::MAX_PAYLOAD_SIZE;
			i++;
			if(size < 0) return -1;
		}
	}
	return 0;
}

void MF_SocketManager::sendData(vector<u_char *> *chkPkt, u_int size){
	MF_Log::mf_log(MF_DEBUG, "MF_SocketManager: sendData send data to API layer");
	u_int n;
	u_int reminSize = size;
	for (n = 0; n < chkPkt->size(); n++) {
		if (n%MAX_CHUNK_SIZE == 0) {
			if (reminSize <= MF_PacketSupport::MAX_PAYLOAD_SIZE - MF_PacketSupport::HIGH_HEADER_LEN) {
				send(mDataSock, MF_PacketSupport::getData((*chkPkt)[n], true), reminSize, 0);
				return;
			}
			else if (reminSize > MF_PacketSupport::MAX_PAYLOAD_SIZE - MF_PacketSupport::HIGH_HEADER_LEN) {
				send(mDataSock, MF_PacketSupport::getData((*chkPkt)[n], true), MF_PacketSupport::MAX_PAYLOAD_SIZE - MF_PacketSupport::HIGH_HEADER_LEN, 0);
				reminSize -= MF_PacketSupport::MAX_PAYLOAD_SIZE - MF_PacketSupport::HIGH_HEADER_LEN;
			}
		} else {
			if (reminSize <= (u_int)MF_PacketSupport::MAX_PAYLOAD_SIZE) {
				send(mDataSock, MF_PacketSupport::getData((*chkPkt)[n], false), reminSize, 0);
				return;
			}
			else if (reminSize > (u_int)MF_PacketSupport::MAX_PAYLOAD_SIZE) {
				send(mDataSock, MF_PacketSupport::getData((*chkPkt)[n], false), MF_PacketSupport::MAX_PAYLOAD_SIZE, 0);
				reminSize -= MF_PacketSupport::MAX_PAYLOAD_SIZE;
			}
		}
	}
}

void MF_SocketManager::sendMessage(vector<u_char *> *v, u_int UID, u_int reqID, u_int size, u_int dstGUID, mfflag_t opts){
	struct MsgSendReply reply;
	reply.UID = UID;
	reply.reqID = reqID;
	reply.type = MSG_TYPE_SEND_REPLY;
	reply.retValue = 0;
	ms->send((char *)(&reply), sizeof(struct MsgSendReply), &cliAddr);
	MF_Log::mf_log(MF_TIME, "MF_SocketManager:sendMessage SEND_TIME request number [%lu] REPLIED", reqID);
	if(recvData(v, size)){
		MF_Log::mf_log(MF_ERROR, "MF_SocketManager:sendMessage error transfering from API layer");
		return;
	}
	struct MsgSendComplete msc;
	msc.UID = UID;
	msc.reqID = reqID;
	msc.type = MSG_TYPE_SEND_COMPLETE;
	ms->send((char *)(&msc), sizeof(struct MsgSendComplete), &cliAddr);
	//pass data to transport
	MF_EvDownAPI *newEvent;
	if(GUIDs.empty()){
		MF_Log::mf_log(MF_DEBUG, "MF_SocketManager:sendMessage passing new message to transport. Size=%u, Src=%d, Dst=%d, opts=%lu", size, 0, dstGUID, opts);
		newEvent = new MF_EvDownAPI(trnsp, v, size, 0, dstGUID, opts);
	}
	else{
		MF_Log::mf_log(MF_DEBUG, "MF_SocketManager:sendMessage passing new message to transport. Size=%u, Src=%d, Dst=%d, opts=%lu", size, GUIDs.front(), dstGUID, opts);
		newEvent = new MF_EvDownAPI(trnsp, v, size, GUIDs.front(), dstGUID, opts);
	}
	eq->add(newEvent);
}

void MF_SocketManager::sendRequest(MF_APISend *event){
	MF_Log::mf_log(MF_DEBUG, "MF_SocketManager:sendRequest handle send request from API layer");
	vector<u_char *> *v = new vector<u_char *>();
	//TODO use of HIGH_HEADER_LEN bad
	if(pendingSend.empty()){
		if(buffer.getVectorBySize(v, event->getSize() + MF_PacketSupport::HIGH_HEADER_LEN*MF_BasicTransport::computeMsgChkCnt(event->getSize()))){
			MF_Log::mf_log(MF_WARN, "MF_SocketManager:sendRequest not enough space in buffer");
			//TODO: introduce flag to request non blocking send
			//if non blocking
			if(false){
				struct MsgSendReply reply;
				reply.UID = event->getUID();
				reply.reqID = event->getReqID();
				reply.type = MSG_TYPE_SEND_REPLY;
				reply.retValue = N_SEND_BUF;
				ms->send((char *)(&reply), sizeof(struct MsgSendReply), &cliAddr);
			}
			else{
				MF_Log::mf_log(MF_DEBUG, "MF_SocketManager:sendRequest the message will wait until there is memory available");
				pendingSend.push(event);
			}
		}
		else{
			sendMessage(v, event->getUID(), event->getReqID(), event->getSize(), event->getDstGUID(), event->getOpts());
		}
	}
	else{
		pendingSend.push(event);
	}
}

//Adjust the fact that we are now passing the size of what is passed in bytes and not number of packets
void MF_SocketManager::recvRequest(MF_APIRecv *event){
	MF_Log::mf_log(MF_DEBUG, "MF_SocketManager:recvRequest handle receive request from API");
	struct MsgRecvReply rcR;
	if(!pendingRec.empty()){
		MF_Log::mf_log(MF_DEBUG, "MF_SocketManager::recvRequest other recv requests are already waiting");
		if(event->getBlk()){
			MF_Log::mf_log(MF_DEBUG, "MF_SocketManager::recvRequest the request is blocking so will wait");
			MF_APIRecv *newEvent = new MF_APIRecv();
			newEvent->setBlk(event->getBlk());
			newEvent->setReqID(event->getReqID());
			newEvent->setSize(event->getSize());
			newEvent->setUID(event->getUID());
			pendingRec.push_back(newEvent);
		}
		else{
			MF_Log::mf_log(MF_DEBUG, "MF_SocketManager::recvRequest the request is non blocking so it will not wait");
			rcR.type = MSG_TYPE_RECV_REPLY;
			rcR.UID = ID;
			rcR.retValue = N_RECV_BUF;
			rcR.reqID = event->getReqID();
			rcR.size = 0;
			ms->send((char *)&rcR, sizeof(struct MsgRecvReply), &cliAddr);
		}
	}
	else if(trnsp->dataAvailable()){
		MF_Log::mf_log(MF_DEBUG, "MF_SocketManager::recvRequest data available in transport");
		//check with transport if there is something available
		vector<u_char *> *v = new vector<u_char *>();
		u_int size = trnsp->getData(v, event->getSize());
		if(size){
			MF_Log::mf_log(MF_DEBUG, "MF_SocketManager::recvRequest the buffer is big enough, getting ready to pass data");
			rcR.type = MSG_TYPE_RECV_REPLY;
			rcR.UID = ID;
			rcR.retValue = 0;
			rcR.reqID = event->getReqID();
			//TODO: check this is correct
			rcR.size = size;
			//rcR.size = MF_BasicTransport::computeChunkPktCnt(size + MF_BasicTransport::computeMsgChkCnt(size)*HIGH_HEADER_LEN);
			RoutingHeader *route;
			route = new RoutingHeader(MF_PacketSupport::getNetworkHeaderPtr(v->front()));
			rcR.sGUID = route->getSrcGUID().to_int();
			ms->send((char *)&rcR, sizeof(struct MsgRecvReply), &cliAddr);
			sendData(v, size);
			buffer.putVector(v);
		}
		else{
			MF_Log::mf_log(MF_DEBUG, "MF_SocketManager::recvRequest the buffer is not big enough, reply will trigger an error");
			rcR.type = MSG_TYPE_RECV_REPLY;
			rcR.UID = ID;
			rcR.retValue = N_RECV_BUF;
			rcR.reqID = event->getReqID();
			rcR.size = 0;
			ms->send((char *)&rcR, sizeof(struct MsgRecvReply), &cliAddr);
		}
	}
	else if(event->getBlk()){
		MF_Log::mf_log(MF_DEBUG, "MF_SocketManager::recvRequest no data available and the request is blocking so will wait");
		MF_APIRecv *newEvent = new MF_APIRecv();
		newEvent->setBlk(event->getBlk());
		newEvent->setReqID(event->getReqID());
		newEvent->setSize(event->getSize());
		newEvent->setUID(event->getUID());
		pendingRec.push_back(newEvent);
	}
	else{
		MF_Log::mf_log(MF_DEBUG, "MF_SocketManager::recvRequest no data available and the request is not blocking so will not wait");
		rcR.type = MSG_TYPE_RECV_REPLY;
		rcR.UID = ID;
		rcR.retValue = N_RECV_BUF;
		rcR.reqID = event->getReqID();
		rcR.size = 0;
		ms->send((char *)&rcR, sizeof(struct MsgRecvReply), &cliAddr);
	}
}

void MF_SocketManager::sendGetMessage(vector<u_char *> *v, u_int UID, u_int reqID, u_int size, u_int dstGUID, mfflag_t opts){
	struct MsgGetReply reply;
	reply.UID = UID;
	reply.reqID = reqID;
	reply.type = MSG_TYPE_GET_SEND_REPLY;
	reply.retValue = 0;
	ms->send((char *)(&reply), sizeof(struct MsgGetReply), &cliAddr);
	MF_Log::mf_log(MF_TIME, "MF_SocketManager:sendGetMessage SEND_TIME request number [%lu] REPLIED", reqID);
	if(recvData(v, size)){
		MF_Log::mf_log(MF_ERROR, "MF_SocketManager:sendGetMessage error transfering from API layer");
		return;
	}
	struct MsgGetSent msc;
	msc.UID = UID;
	msc.reqID = reqID;
	msc.type = MSG_TYPE_GET_SEND_COMPLETE;
	ms->send((char *)(&msc), sizeof(struct MsgGetSent), &cliAddr);
	//Add get to pendingGet
	pendingGet.insert(make_pair(dstGUID, reqID));
	//pass data to transport
	MF_EvDownAPI *newEvent;
	if(GUIDs.empty()){
		MF_Log::mf_log(MF_DEBUG, "MF_SocketManager:sendGetMessage passing new message to transport. Size=%u, Src=%d, Dst=%d, opts=%lu", size, 0, dstGUID, opts);
		newEvent = new MF_EvDownAPI(trnsp, v, size, 0, dstGUID, setMFFlag(opts, MF_CONTENT_REQUEST));
	}
	else{
		MF_Log::mf_log(MF_DEBUG, "MF_SocketManager:sendGetMessage passing new message to transport. Size=%u, Src=%d, Dst=%d, opts=%lu", size, GUIDs.front(), dstGUID, opts);
		newEvent = new MF_EvDownAPI(trnsp, v, size, GUIDs.front(), dstGUID, setMFFlag(opts, MF_CONTENT_REQUEST));
	}
	eq->add(newEvent);
}

void MF_SocketManager::getRequest(MF_APIGet *event){
	//TODO: if space available for sending request sends, otherwise returns error
	MF_Log::mf_log(MF_DEBUG, "MF_SocketManager:getRequest handle get request from API layer");
	vector<u_char *> *v = new vector<u_char *>();
	//TODO use of HIGH_HEADER_LEN bad
	if(buffer.getVectorBySize(v, event->getSize() + MF_PacketSupport::HIGH_HEADER_LEN*MF_BasicTransport::computeMsgChkCnt(event->getSize()))){
			MF_Log::mf_log(MF_WARN, "MF_SocketManager:getRequest not enough space in buffer");
			//send back error to the API
		struct MsgGetReply reply;
		reply.UID = event->getUID();
		reply.reqID = event->getReqID();
		reply.type = MSG_TYPE_GET_SEND_REPLY;
		reply.retValue = N_SEND_BUF;
		ms->send((char *)(&reply), sizeof(struct MsgGetReply), &cliAddr);
	}
	else{
		sendGetMessage(v, event->getUID(), event->getReqID(), event->getSize(), event->getDstGUID(), event->getOpts());
	}
}

void MF_SocketManager::replyDoGetRequest(MF_APIDoGet *event){
	MF_Log::mf_log(MF_DEBUG, "MF_SocketManager:replyDoGetRequest");
	struct MsgDoGetReply rcR;
	//check with transport if there is something available
	vector<u_char *> *v = pendingGetRequest.front();
	pendingGetRequest.pop();
	u_int size = pendingGetRequestSizes.front();
	pendingGetRequestSizes.pop();
	rcR.type = MSG_TYPE_DO_GET_REPLY;
	rcR.UID = event->getUID();
	rcR.retValue = 0;
	rcR.reqID = event->getReqID();
	lastGetRequest++;
	rcR.getID = lastGetRequest;
	rcR.size = size;
	//rcR.size = MF_BasicTransport::computeChunkPktCnt(size + MF_BasicTransport::computeMsgChkCnt(size)*HIGH_HEADER_LEN);
	RoutingHeader *route;
	route = new RoutingHeader(MF_PacketSupport::getNetworkHeaderPtr(v->front()));
	rcR.srcGUID = route->getSrcGUID().to_int();
	rcR.dstGUID = route->getDstGUID().to_int();
	ms->send((char *)&rcR, sizeof(struct MsgDoGetReply), &cliAddr);
	sendData(v, size);
	buffer.putVector(v);
}

void MF_SocketManager::doGetRequest(MF_APIDoGet *event){
	//TODO: very similar a recv, if there are some messages pending it will push them up, otherwise will start waiting
	MF_Log::mf_log(MF_DEBUG, "MF_SocketManager:doGetRequest handle do_get request from API");
	if(!pendingGetRequest.empty()){
		replyDoGetRequest(event);
	}
	else{
		MF_APIDoGet *tempEvent = new MF_APIDoGet();
		tempEvent->setReqID(event->getReqID());
		tempEvent->setSize(event->getSize());
		tempEvent->setUID(event->getUID());
		pendingDoGet.push(event);
	}
}

void MF_SocketManager::sendGetResponseMessage(vector<u_char *> *v, u_int UID, u_int reqID, u_int size, u_int srcGUID, u_int dstGUID, mfflag_t opts){
	struct MsgGetResponseReply reply;
	reply.UID = UID;
	reply.reqID = reqID;
	reply.type = MSG_TYPE_GET_RESPONSE_REPLY;
	reply.retValue = 0;
	ms->send((char *)(&reply), sizeof(struct MsgGetResponseReply), &cliAddr);
	MF_Log::mf_log(MF_TIME, "MF_SocketManager:sendGetResponseMessage GET_RESPONSE_TIME request number [%lu] REPLIED", reqID);
	if(recvData(v, size)){
		MF_Log::mf_log(MF_ERROR, "MF_SocketManager:sendGetResponseMessage error transfering from API layer");
		return;
	}
	struct MsgGetResponseComplete msc;
	msc.UID = UID;
	msc.reqID = reqID;
	msc.type = MSG_TYPE_GET_RESPONSE_COMPLETE;
	ms->send((char *)(&msc), sizeof(struct MsgGetResponseComplete), &cliAddr);
	//pass data to transport
	MF_EvDownAPI *newEvent;
	//NOTE: now srcGUID and dstGUID are getting swapped because they still are the original src and dst of the get request
	//TODO: make this swapping more clear
	MF_Log::mf_log(MF_DEBUG, "MF_SocketManager:sendGetResponseMessage passing new message to transport. Size=%u, Src=%d, Dst=%d, opts=%lu", size, dstGUID, srcGUID, opts);
	newEvent = new MF_EvDownAPI(trnsp, v, size, dstGUID, srcGUID, setMFFlag(opts, MF_CONTENT_RESPONSE));
	eq->add(newEvent);
}

void MF_SocketManager::getResponseRequest(MF_APIGetResponse *event){
	//When a do_get is pushed up, an entry will be created in the map for when the responsed is actually pushed
	//Modify from send
	MF_Log::mf_log(MF_DEBUG, "MF_SocketManager:getResponseRequest handle send request from API layer");
	vector<u_char *> *v = new vector<u_char *>();
	//TODO use of HIGH_HEADER_LEN bad
	if(pendingGetResponse.empty()){
		if(buffer.getVectorBySize(v, event->getSize() + MF_PacketSupport::HIGH_HEADER_LEN*MF_BasicTransport::computeMsgChkCnt(event->getSize()))){
			MF_Log::mf_log(MF_WARN, "MF_SocketManager:getResponseRequest not enough space in buffer");
			pendingGetResponse.push(event);
		}
		else{
			sendGetResponseMessage(v, event->getUID(), event->getReqID(), event->getSize(), event->getSrcGUID(), event->getDstGUID(), event->getOpts());
		}
	}
	else{
		pendingGetResponse.push(event);
	}
}

void MF_SocketManager::closeRequest(){
	MF_Log::mf_log(MF_DEBUG, "MF_SocketManager:closeRequest handle close request from API");
	//TODO:clean things if there is something to clean
	st->removeManager(ID);
	delete trnsp;
	buffer.dePopulate();
	char tempName[128];
	memset(tempName, 0, 128);
	sprintf(tempName, "%s%010x", SERVER_DATA_PATH, ID);
	unlink(tempName);
}


MF_Transport *MF_SocketManager::getTransportByTID(u_int tid){
	MF_Log::mf_log(MF_DEBUG, "MF_SocketManager:getTransoirtByTID return transport object given TID %u", tid);
	//TODO:at the moment only one transport but it should check in the future
	return trnsp;
}

void MF_SocketManager::basicMsgAvailable(MF_EvUpTransp *event){
	if(!pendingRec.empty()){
		MF_Log::mf_log(MF_DEBUG, "MF_SocketManager::msgAvailable there are pending recv requests");
		vector<MF_APIRecv *>::iterator it = pendingRec.begin();
		bool cont = true;
		while (it!=pendingRec.end() && cont) {
			MF_Log::mf_log(MF_DEBUG, "MF_SocketManager::msgAvailable checking among pending recv request %u", (*it)->getReqID());
			if((*it)->getSize() < event->getSize()){
				MF_Log::mf_log(MF_DEBUG, "MF_SocketManager::msgAvailable request %u has buffer of size %u too small for message of size %u", (*it)->getReqID(), (*it)->getSize(), event->getSize());
				it++;
			}
			else{
				MF_Log::mf_log(MF_DEBUG, "MF_SocketManager::msgAvailable sending data to request %u", (*it)->getReqID());
				struct MsgRecvReply rcR;
				cont = false;
				MF_APIRecv *temp = (*it);
				pendingRec.erase(it);
				rcR.type = MSG_TYPE_RECV_REPLY;
				rcR.retValue = 0;
				rcR.UID = ID;
				rcR.reqID = temp->getReqID();
				vector<u_char *> *v = new vector<u_char *>();
				u_int avail = trnsp->getData(v, temp->getSize());
				if(avail>0){
					//TODO: change this to get the info from the transport and not the packet
					RoutingHeader *route;
					route = new RoutingHeader(MF_PacketSupport::getNetworkHeaderPtr(v->front()));
					rcR.sGUID = route->getSrcGUID().to_int();
					//TODO: check this is correct
					rcR.size = avail;
					//rcR.size = MF_BasicTransport::computeChunkPktCnt(avail+MF_BasicTransport::computeMsgChkCnt(avail)*HIGH_HEADER_LEN);
					MF_Log::mf_log(MF_INFO, "MF_SocketManager::msgAvailable replying to request %u with %u bytes of data", (*it)->getReqID(), avail);
					ms->send((char *)&rcR, sizeof(struct MsgRecvReply), &cliAddr);
					MF_Log::mf_log(MF_TIME, "MF_Main:handleUpTransp RECV_TIME REPLIED");
					sendData(v, avail);
					buffer.putVector(v);
				}
				else {
					MF_Log::mf_log(MF_DEBUG, "MF_SocketManager::msgAvailable cannot send message because returned size from transport is 0");
				}
				
			}
		}
		if(cont==false){
			//MF_Log::mf_log(MF_DEBUG, "MF_SocketManager::msgAvailable: No waiting receive has enough buffer");
		}
	}
	else{
		MF_Log::mf_log(MF_INFO, "MF_SocketManager::msgAvailable no recv requests pending for UID %u", ID);
	}
}

void MF_SocketManager::contentMsgAvailable(MF_EvUpTransp *event){
	//TODO: check, if it's a content request, it checks if there is an available do_get pending and pushes up otherwise waits.
	//In any case it takes it back from the transport
	
	vector<u_char *> *v = new vector<u_char *>();
	//TODO 99999999 is just a place holder to make sure the msg is passed up from transport
	u_int avail = trnsp->getData(v, 999999999);
	//TODO ho dimenticato che questo può essere una risposta per la get mandata, ma anche una richiesta da passare al do_get se ce n'e'
	if(avail>0){
		RoutingHeader *route;
		route = new RoutingHeader(MF_PacketSupport::getNetworkHeaderPtr(v->front()));
		if(route->getServiceID().isContentResponse()){
			MF_Log::mf_log(MF_DEBUG, "MF_SocketManager::contentMsgAvailable received a new content response");
			u_int srcGUID = (u_int)route->getSrcGUID().to_int();
			map<unsigned int, unsigned int>::iterator it = pendingGet.find(srcGUID);
			if (it!=pendingGet.end()) {
				struct MsgGetComplete mgc;
				mgc.type = MSG_TYPE_GET_SEND_RECV;
				mgc.UID = ID;
				mgc.reqID = it->second;
				mgc.retValue = 0;
				mgc.size = avail;
				ms->send((char *)&mgc, sizeof(struct MsgGetComplete), &cliAddr);
				sendData(v, avail);
				buffer.putVector(v);
				pendingGet.erase(srcGUID);
			}
			else{
				MF_Log::mf_log(MF_WARN, "MF_SocketManager::contentMsgAvailable no pending request for guid %u", srcGUID);
				buffer.putVector(v);
			}
		}
		else if(route->getServiceID().isContentRequest()){
			MF_Log::mf_log(MF_DEBUG, "MF_SocketManager::contentMsgAvailable received a new content request");
			pendingGetRequest.push(v);
			pendingGetRequestSizes.push(avail);
			if(!pendingDoGet.empty()){
				MF_APIDoGet *doGetEvent = pendingDoGet.front();
				pendingDoGet.pop();
				replyDoGetRequest(doGetEvent);
			}
		}
		else{
			MF_Log::mf_log(MF_WARN, "MF_SocketManager::contentMsgAvailable I am working on content but the message was not a content request or response");
			buffer.putVector(v);
		}
	}
	else {
		MF_Log::mf_log(MF_ERROR, "MF_SocketManager::contentMsgAvailable cannot send message because returned size from transport is 0");
		buffer.putVector(v);
	}
}

//wrong behaviour it should send some error message
void MF_SocketManager::msgAvailable(MF_EvUpTransp *event){
	MF_Log::mf_log(MF_DEBUG, "MF_SocketManager::msgAvailable new message available to be delivered to API from socket %u", ID);
	if(type == 1){
		basicMsgAvailable(event);
	}
	else{
		contentMsgAvailable(event);
	}
}

void MF_SocketManager::releaseChunk(MF_ChunkInfo *ci){
	MF_Log::mf_log(MF_DEBUG, "MF_SocketManager::releaseChunk release message sent");
	vector<u_char *> *v = ci->getPacketList();
	if(v!=NULL){
		buffer.putVector(v);
		//Check if there are pending sends to continue
		newSpaceAvail();
	}
	else{
		MF_Log::mf_log(MF_WARN, "MF_SocketManager::releaseChunk releasing empty buffer");
	}
}

//TODO: do we want to do this operation in a new event in the stack (think so)
void MF_SocketManager::newSpaceAvail(){
	MF_Log::mf_log(MF_DEBUG, "MF_SocketManager:newSpaceAvail some buffer has been freed up; checking if there are pending send requests");
	vector<u_char *> *v = new vector<u_char *>();
	if(type == 1){
		if(!pendingSend.empty()){
			MF_APISend *event = pendingSend.front();
			//TODO use of HIGH_HEADER_LEN bad
			if(buffer.getVectorBySize(v, event->getSize() + MF_PacketSupport::HIGH_HEADER_LEN*MF_BasicTransport::computeMsgChkCnt(event->getSize()))){
				MF_Log::mf_log(MF_DEBUG, "MF_SocketManager:sendRequest the message will wait until there is memory available");
			}
			else{
				pendingSend.pop();
				sendMessage(v, event->getUID(), event->getReqID(), event->getSize(), event->getDstGUID(), event->getOpts());
			}
		}
	}
	else{
		if(!pendingGetResponse.empty()){
			MF_APIGetResponse *event = pendingGetResponse.front();
			//TODO use of HIGH_HEADER_LEN bad
			if(buffer.getVectorBySize(v, event->getSize() + MF_PacketSupport::HIGH_HEADER_LEN*MF_BasicTransport::computeMsgChkCnt(event->getSize()))){
				MF_Log::mf_log(MF_DEBUG, "MF_SocketManager:sendRequest the message will wait until there is memory available");
			}
			else{
				pendingGetResponse.pop();
				sendGetResponseMessage(v, event->getUID(), event->getReqID(), event->getSize(), event->getSrcGUID(), event->getDstGUID(), event->getOpts());
			}
		}
	}
}
